//modified by Jing Yang after 7/6/2001

#include "okcdata.H"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "macros.H"
#include "brush.H"
#include <math.h>
#include <search.h>
#include "CScreenDistortion.H"
#include "XmdvTool.H"
#include "ARR.H"


static const char *read_till(FILE *fp, const char *delimiters="\n", char *which_delimiter=0)
{
    static char str[500];
    int k=0;
    while (1) {
        char c=fgetc(fp);
        char *ptr = strchr(delimiters, c);
        if (ptr) {
            str[k] = '\0';
            if (which_delimiter) *which_delimiter = *ptr;
            break;
        }
        str[k++] = c;
    }

	/*if ( strchr(str, ' ') )
	{
	    char str1[500];
		sprintf(str1, "{%s}", str);
		strcpy(str, str1);
	}*/
    return str;
}

// Read in data file: Format is:
// number_of_dimensions number_of_samples (int int)
// dimension_name for each dimension (string)
// minimum maximum cardinality for each dimension (double double int)
// data_samples (doubles)
//
void okcdata::Read(const char *fname)
{
    FILE *fp = fopen(fname, "r");
    if(fp == NULL)
    {
        fprintf(stderr,"Data file %s doesn't exist\n", fname);
        return;
    }

    datafile = strdup(fname);

    // Get number of dimensions, and number of records
    //
    if (fscanf(fp, "%d %d\n", &all_dims, &data_size) <= 0)  {
        printf("Error reading first line of file %s!\n", fname);
        printf("Try loading another file!\n");
        all_dims = 0;
        data_size = 0;
    }

    if (data_size < 0)  {
        printf("Number of data items is negative (%d)!\n", data_size);
        printf("Try loading another file!\n");
        all_dims = 0;
        data_size = 0;
    }

    // Read in names of dimensions, and if necessary, nominal
    // encodings of values.
    //
    all_names.SetLen(all_dims);
    all_nominal_fields.SetLen(all_dims);

    int i;
    for (i=0; i < all_dims; i++)
    {
        char endchar;
        const char *dname = read_till(fp,"(\n",&endchar);
        all_names[i] = new char[strlen(dname)+4];
        strcpy(all_names[i],dname);

        NominalField *ndata = 0;
        if (endchar == '(') 
		{
/*            ndata = new NominalField;
            ndata->range_min = atoi(read_till(fp,",",&endchar));
            ndata->range_max = atoi(read_till(fp,"),\n",&endchar));
            ndata->values.SetLen(ndata->range_max-ndata->range_min+1);

            int n;
            for (n=0;n<ndata->values.Len();n++) {
                ndata->values[n] = 0;
            }

            if (endchar != ')') 
			{
                for (n=ndata->range_min; n<=ndata->range_max; n++) {
                    const char *str = read_till(fp,",)\n",&endchar);
                    ndata->values[n-ndata->range_min] = strdup(str);
                }
            }
*/
            read_till(fp,"\n");
        }

        all_nominal_fields[i] = ndata;
    }

    all_dim_min.SetLen(all_dims);
    all_dim_max.SetLen(all_dims);
    cg_dim_min.SetLen(all_dims);
    cg_dim_max.SetLen(all_dims);
    all_cardinality.SetLen(all_dims);
    on.SetLen(all_dims);

    for (i=0;i < all_dims;i++)
    {
        if (fscanf(fp, "%lf %lf %d", &all_dim_min[i], &all_dim_max[i], &all_cardinality[i]) !=3) {
            printf("failed in reading dimension ranges.\n");
            exit(-1);
        }

		cg_dim_min[i] = all_dim_min[i];
		cg_dim_max[i] = all_dim_max[i];

        // Need to handle case when min = max (separate them a bit)
        //
        if (all_dim_min[i] == all_dim_max[i]) {
            if (all_dim_min[i] == 0.0)  {
                all_dim_min[i] = -1.0;
                all_dim_max[i] = 1.0;
            } else {
                if (all_dim_min[i] < 0.0) {
                    all_dim_min[i] = 2. * all_dim_min[i];
                    all_dim_max[i] = 0.0;
                } else {
                    all_dim_max[i] = 2. * all_dim_min[i];
                    all_dim_min[i] = 0.0;
                }
            }
        } else {

            // Add a bit of space to the bounds to avoid boundary problems
            //
            const double SPACE=.05; // extra space for bounds to avoid landing on edges

            double dim_diff = (all_dim_max[i] - all_dim_min[i]) * SPACE;
            all_dim_max[i] += dim_diff;
            all_dim_min[i] -= dim_diff;
        }

        on[i] = true;
    }

    data_buf.SetLen(all_dims*data_size);

    int k=0;
    for (i=0;i < data_size;i++) {
        for (int j = 0;j < all_dims;j++,k++) {
            if (fscanf(fp, "%lf", &data_buf[k]) != 1) {
                printf("failed in reading data, line %d, item %d.\n", i, j);
            }
        }
    }

    fclose(fp);

	mark.SetLen(data_size);
    color.SetLen(data_size);

    names.SetLen(all_dims);
    dim_min.SetLen(all_dims);
    dim_max.SetLen(all_dims);
    nominal_fields.SetLen(all_dims);
    org_dim_min.SetLen(all_dims);
    org_dim_max.SetLen(all_dims);
    cardinality.SetLen(all_dims);
	//used for reordering dimensions
    map_file_to_buf.SetLen(all_dims);
    map_buf_to_file.SetLen(all_dims);
	for ( i = 0; i < all_dims; i++ )
	{
		map_file_to_buf[i] = i;
		map_buf_to_file[i] = i;
	}


	//used for glyph placement
	basex = new double[data_size];
	basey = new double[data_size];

	Placement_Original_Glyph();

    SetFieldOnOff(-1,true);
}

//read from file for reordering.
void okcdata::Read2()
{
    FILE *fp = fopen(datafile, "r");
    if(fp == NULL)
    {
        fprintf(stderr,"Data file %s doesn't exist\n", datafile);
        return;
    }

    // Get number of dimensions, and number of records
    //
    int all_dims2, data_size2;
	if (fscanf(fp, "%d %d\n", &all_dims2, &data_size2) <= 0)  {
        printf("Error reading first line of file %s!\n", datafile);
        printf("Try loading another file!\n");
        all_dims = 0;
        data_size = 0;
    }

	if ( ( all_dims2 != all_dims ) || ( data_size2 != data_size ) )
		return;

    // Read in names of dimensions, and if necessary, nominal
    // encodings of values.
    //
    int i, loopi;
    for (loopi=0; loopi < all_dims; loopi++)
    {
        i = map_file_to_buf[loopi];
		char endchar;
        const char *dname = read_till(fp,"(\n",&endchar);
        all_names[i] = new char[strlen(dname)+4];
        strcpy(all_names[i],dname);

        NominalField *ndata = 0;
        if (endchar == '(') {
            ndata = new NominalField;
            ndata->range_min = atoi(read_till(fp,"),\n",&endchar));
            ndata->range_max = atoi(read_till(fp,"),\n",&endchar));
            ndata->values.SetLen(ndata->range_max-ndata->range_min+1);

            int n;
            for (n=0;n<ndata->values.Len();n++) {
                ndata->values[n] = 0;
            }

            if (endchar != ')') {
                for (n=ndata->range_min; n<=ndata->range_max; n++) {
                    const char *str = read_till(fp,",)\n",&endchar);
                    ndata->values[n-ndata->range_min] = strdup(str);
                }
            }

            read_till(fp,"\n");
        }

        all_nominal_fields[i] = ndata;
    }

    for (loopi=0; loopi < all_dims; loopi++)
    {
        i = map_file_to_buf[loopi];
        if (fscanf(fp, "%lf %lf %d", &all_dim_min[i], &all_dim_max[i], &all_cardinality[i]) !=3) {
            printf("failed in reading dimension ranges.\n");
            exit(-1);
        }
		cg_dim_min[i] = all_dim_min[i];
		cg_dim_max[i] = all_dim_max[i];

        // Need to handle case when min = max (separate them a bit)
        //
        if (all_dim_min[i] == all_dim_max[i]) {
            if (all_dim_min[i] == 0.0)  {
                all_dim_min[i] = -1.0;
                all_dim_max[i] = 1.0;
            } else {
                if (all_dim_min[i] < 0.0) {
                    all_dim_min[i] = 2. * all_dim_min[i];
                    all_dim_max[i] = 0.0;
                } else {
                    all_dim_max[i] = 2. * all_dim_min[i];
                    all_dim_min[i] = 0.0;
                }
            }
        } else {

            // Add a bit of space to the bounds to avoid boundary problems
            //
            const double SPACE=.05; // extra space for bounds to avoid landing on edges

            double dim_diff = (all_dim_max[i] - all_dim_min[i]) * SPACE;
            all_dim_max[i] += dim_diff;
            all_dim_min[i] -= dim_diff;
        }

    }

    int base, k;
    for (i=0;i < data_size;i++) {
		base = i * all_dims;
        for (int j = 0;j < all_dims;j++) {
	        k = base + map_file_to_buf[j];
            if (fscanf(fp, "%lf", &data_buf[k]) != 1) {
                printf("failed in reading data, line %d, item %d.\n", i, j);
            }
        }
    }


    fclose(fp);

}


void okcdata::WriteBrushed(const char *fname)
{
	if ( strcmp ( fname, "" ) == 0 )
		return;
	int i, j;

    int count = 0;
    
    ARR<double> data;
	ARR<double> max;
	ARR<double> min;

    data.SetLen(dims);
    max.SetLen(dims);
    min.SetLen(dims);

    bool brushed;

    for (i=0;i < dims;i++)
    {
        max[i] = curr_brush->Min(i);
		min[i] = curr_brush->Max(i);
	}

	for ( i=0; i<data_size; i++) 
	{
        GetData(data,i);
		brushed = true;
		for ( j =0; j < dims; j++) 
		{
			if ( data[j] < curr_brush->Min(j)) 
				brushed = false;
			if ( data[j] > curr_brush->Max(j)) 
				brushed = false;
	    }
		if ( brushed )
		{
			for ( j = 0; j < dims; j++ )
			{
				if ( data[j] > max[j] )
					max[j] = data[j];
				if ( data[j] < min[j] )
					min[j] = data[j];
			}
			count ++;
		}
    }

	
	
	FILE *fp = fopen(fname, "w");

    // Get number of dimensions, and number of records
    //
    fprintf(fp, "%d %d\n", dims, count);

    // Read in names of dimensions, and if necessary, nominal
    // encodings of values.
    //
    for (i=0; i < dims; i++)
    {
	    fprintf(fp, "%s\n", names[i]);
    }

    for (i=0;i < dims;i++)
    {
        fprintf(fp, "%lf %lf %d\n", min[i], max[i], cardinality[i]);
	}

	for ( i=0; i<data_size; i++) 
	{
        GetData(data,i);
		brushed = true;
		for ( j =0; j < dims; j++) 
		{
			if ( data[j] < curr_brush->Min(j)) 
				brushed = false;
			if ( data[j] > curr_brush->Max(j)) 
				brushed = false;
	    }
		if ( brushed )
		{
			for ( j = 0; j < dims; j++ )
			{
				fprintf(fp, "%lf ", data[j]);
			}
			fprintf(fp, "\n");
		}
    }

    fclose(fp);
}

void okcdata::GetData(ARR<double> &buf, int line)
{
    ARR<double> all_data(all_dims);
    
    int j=0;
    int base = all_dims*line;
    for (int i=0;i<all_dims;i++)
            all_data[j++] = data_buf[base + i];
    MapXData(all_data, buf); 
} 


void okcdata::GetDataWithDimClusterExtends(ARR<double> &mean, ARR<double> &max, ARR<double> &min, int line)
{
    ARR<double> all_data(all_dims);
    
    int j=0;
    int base = all_dims*line;
    for (int i=0;i<all_dims;i++)
            all_data[j++] = data_buf[base + i];
    MapDataWithDimClusterExtends(all_data, mean, max, min); 
} 


#include "brush.H"

void okcdata::SaveBrushes()
{
	for (int br=0; br<MAXBRUSH; br++)
    {
        Brush *all_brush = &all_brushes[br];
        Brush *brush = &brushes[br];
        ReverseMapData(brush->pos, all_brush->pos);
        ReverseMapData(brush->size, all_brush->size);
        ReverseMapData(brush->bound_offset, all_brush->bound_offset);
	}
}


void okcdata::CountDims()
{
    dims = 0;
    for (int i=0;i<all_dims;i++)
        if (on[i]) 
        {
            dims++;
        }
}

void okcdata::SetFieldOnOff(int which, bool state)
{
    if (which>=0) 
    {
        SaveBrushes();
        on[which] = state;
    }

    dims = 0;
    for (int i=0;i<all_dims;i++)
        if (on[i]) 
        {
            dims++;
        }

    one_over_dims = 1.0/dims;

    MapXData(all_dim_max, dim_max);
    MapXData(all_dim_min, dim_min);
    MapXData(all_dim_max, org_dim_max);
    MapXData(all_dim_min, org_dim_min);
    MapXData(all_cardinality, cardinality);
    MapXData(all_names, names);
    MapXData(all_nominal_fields, nominal_fields);

    if (which>=0) 
    {
	    for (int br=0; br<MAXBRUSH; br++)
        {
            Brush *all_brush = &all_brushes[br];
            Brush *brush = &brushes[br];
            MapXData(all_brush->pos, brush->pos);
            MapXData(all_brush->size, brush->size);
            MapXData(all_brush->bound_offset, brush->bound_offset);
	    }
    }
	screen_distortion->Reset();

    if ( which >= 0 && dim_reducer != NULL )
    {
        delete dim_reducer;
        dim_reducer = NULL;
    }
}

void okcdata::RemapAllData()
{
    dim_max.SetLen(dims);
    dim_min.SetLen(dims);
    org_dim_max.SetLen(dims);
    org_dim_min.SetLen(dims);
    cardinality.SetLen(dims);
    names.SetLen(dims);
    nominal_fields.SetLen(dims);
    MapXData(all_dim_max, dim_max);
    MapXData(all_dim_min, dim_min);
    MapXData(all_dim_max, org_dim_max);
    MapXData(all_dim_min, org_dim_min);
    MapXData(all_cardinality, cardinality);
    MapXData(all_names, names);
    MapXData(all_nominal_fields, nominal_fields);

	for (int br=0; br<MAXBRUSH; br++)
    {
        Brush *all_brush = &all_brushes[br];
        Brush *brush = &brushes[br];
        brush->pos.SetLen(dims);
        brush->size.SetLen(dims);
        brush->bound_offset.SetLen(dims);
        MapXData(all_brush->pos, brush->pos);
        MapXData(all_brush->size, brush->size);
        MapXData(all_brush->bound_offset, brush->bound_offset);
	}
	screen_distortion->Reset();
}


//reorder the dimensions. Which indicate which dimension need to be forward or backward. If forward == 1, 
// then move this dimension forward. Else move this dimension backward.
void okcdata::ReorderDimension(int which, int forward)
{
    SaveBrushes();

	bool temp_bool;
	int temp_int;

	//neighbor indicates that dimension need to be exchange with "which" dimension.
	int neighbor;

	if ( forward == 1 )
	{
		if ( which == 0 )
			neighbor = all_dims-1;
		else
			neighbor = which-1;
	}
	else 
	{
		if ( which == all_dims-1 )
			neighbor = 0;
		else
			neighbor = which+1;
	}

	temp_int = map_buf_to_file[neighbor];
	map_buf_to_file[neighbor] = map_buf_to_file[which];
	map_buf_to_file[which] = temp_int;

	map_file_to_buf[map_buf_to_file[which]] = which;
	map_file_to_buf[map_buf_to_file[neighbor]] = neighbor;

	temp_bool = on[neighbor];
	on[neighbor] = on[which];
	on[which] = temp_bool;

    RemapAllData();

    if ( which >= 0 && dim_reducer != NULL )
    {
        delete dim_reducer;
        dim_reducer = NULL;
    }
}


/*
void okcdata::SetReorderedBrushes()
{
    int br, i, loopi;
    ARR<double> temp_pos;
    ARR<double> temp_size;
    ARR<double> temp_bound_offset;  // boundary offset from min/max
    ARR<bool> temp_on;

	pos.SetLen(all_dims);
	size.SetLen(all_dims);
	temp_bound_offset.SetLen(all_dims);
	temp_on.SetLen(all_dims);

	for ( i = 0; i < all_dims; i++ )
	{
		temp_on[i] = on[i];
	}

	for ( loopi = 0; loopi < all_dims; loopi++ )
	{
		i = 
		temp_on[i] = on[i];
	}

	for (int br=0; br<MAXBRUSH; br++)
    {
        Brush *brush = &all_brushes[br];
		for ( i = 0; i < all_dims; i ++ )
		{
			temp_pos[i] =  brush.pos[i];
			temp_pos[i] =  brush.pos[i];
			temp_pos[i] =  brush.pos[i];
		}

		
	dims = all_dims;
    names.SetLen(all_dims);
    dim_min.SetLen(all_dims);
    dim_max.SetLen(all_dims);
    nominal_fields.SetLen(all_dims);
    org_dim_min.SetLen(all_dims);
    org_dim_max.SetLen(all_dims);

    cardinality.SetLen(all_dims);

    dims = 0;
    for (i=0;i<all_dims;i++)
        if (on[i]) {
            org_dim_min[dims] = dim_min[dims] = all_dim_min[i];
            org_dim_max[dims] = dim_max[dims] = all_dim_max[i];
            cardinality[dims] = all_cardinality[i];
            names[dims] = all_names[i];
            nominal_fields[dims] = all_nominal_fields[i];
            dims++;
        }


    one_over_dims = 1.0/dims;

    for (br=0; br<MAXBRUSH; br++)
    {
        int j = 0;
        for (int i=0;i<MIN2(brushes[br].pos.Len(),on.Len());i++)
            if (on[i]) {
                brushes[br].pos[j] = all_brushes[br].pos[i];
                brushes[br].size[j] = all_brushes[br].size[i];
                brushes[br].bound_offset[j] = all_brushes[br].bound_offset[i];
                j++;
            }
    }
}
*/


const char* NominalField::MapValueToString(int what)
{
    if (what<range_min || what>range_max || values[what-range_min]==0) {
        static char number_string[10];
        sprintf(number_string,"%d",what);
        return number_string;
    }
    return values[what-range_min];
}


void okcdata::MapXData( ARR<double> all_data, ARR<double> &data)
{
    ARR<double> temp(all_dims);
    int i, loopi;
    for (loopi=0; loopi<all_dims; loopi++)
	{
        i = map_file_to_buf[loopi];
        temp[i] = all_data[loopi];
    }

    ARR<double> temp1(all_dims);
    int count = 0;
    for (i=0;i<all_dims;i++)
    {
        if (on[i]) 
        {
            temp1[count] = temp[i];
            count++;
        }
    }

    if ( (dim_reducer != NULL) && (dim_reducer->status == APPLY) )
    {
        dim_reducer->MapData(temp1, data);        
    }
    else
    {
        for (i=0;i<count;i++)
        {
            data[i] = temp1[i];
        }
    }
}



template <class T> void okcdata::MapXData( ARR<T> all_data, ARR<T> &data)
{
    ARR<T> temp(all_dims);
    int i, loopi;
    for (loopi=0; loopi<all_dims; loopi++)
	{
        i = map_file_to_buf[loopi];
        temp[i] = all_data[loopi];
    }

    ARR<T> temp1(all_dims);
    int count = 0;
    for (i=0;i<all_dims;i++)
    {
        if (on[i]) 
        {
            temp1[count] = temp[i];
            count++;
        }
    }

    if ( (dim_reducer != NULL) && (dim_reducer->status == APPLY) )
    {
        dim_reducer->MapData(temp1, data);        
    }
    else
    {
        for (i=0;i<count;i++)
        {
            data[i] = temp1[i];
        }
    }
}




void okcdata::MapDataWithDimClusterExtends(ARR<double> all_data, ARR<double> &mean, ARR<double> &max, ARR<double> &min)
{
    ARR<double> temp(all_dims);
    int i, loopi;
    for (loopi=0; loopi<all_dims; loopi++)
	{
        i = map_file_to_buf[loopi];
        temp[i] = all_data[loopi];
    }

    ARR<double> temp1(all_dims);
    int count = 0;
    for (i=0;i<all_dims;i++)
    {
        if (on[i]) 
        {
            temp1[count] = temp[i];
            count++;
        }
    }

    dim_reducer->MapDataWithDimClusterExtends(temp1, mean, max, min);        
}


/*
template <class T> void okcdata::MapDataWithoutOnOff( ARR<T> all_data, ARR<T> &data)
{
    int i, loopi;
    for (loopi=0; loopi<all_dims; loopi++)
	{
        i = map_file_to_buf[loopi];
        data[i] = all_data[loopi];
    }

}
*/

template <class T> void okcdata::ReverseMapData( ARR<T> data, ARR<T> &all_data)
{
    int i, file_i;
    int count = 0;
    for (i=0;i<all_dims;i++)
    {
        if (on[i]) 
        {
            file_i = map_buf_to_file[i];
            all_data[file_i] = data[count];
            count++;
        }
    }
}

